
from datetime import datetime
from typing import Any

import requests
import pandas as pd
from pandas import DataFrame


class FinnHubStockReader:
    def __init__(self, token: str):
        self._token = token

    def get_all_US_symbols(self) -> list[dict[str, Any]]:
        request_url = f"https://finnhub.io/api/v1/stock/symbol?exchange=US&token={self._token}"
        r = requests.get(request_url)
        response = r.json()
        return response

    def get_stock_candlestick(self, symbol: str, startdate: str, enddate: str) -> DataFrame:
        starttimestamp = int(datetime.strptime(startdate, '%Y-%m-%d').timestamp())
        endtimestamp = int(datetime.strptime(enddate, '%Y-%m-%d').timestamp())
        request_url = 'https://finnhub.io/api/v1/stock/candle?symbol={}&resolution=1&from={}&to={}&token={}'.format(symbol, starttimestamp, endtimestamp, self._token)
        r = requests.get(request_url)
        response = r.json()
        if response['s'] == 'ok':
            df = pd.DataFrame({
                'TimeStamp': [pd.Timestamp(ts, unit='s') for ts in response['t']],
                'High': response['h'],
                'Low': response['l'],
                'Open': response['o'],
                'Close': response['c'],
                'Volume': response['v']
            })
            return df
        else:
            raise Exception('Error: {}'.format(response['s']))




import json
import logging
import sys
from collections import defaultdict
from typing import Union, Any
from os import PathLike
from io import TextIOWrapper
if sys.version_info < (3, 11):
    from typing_extensions import Self
else:
    from typing import Self

from tqdm import tqdm
import numpy as np
import pandas as pd
from nptyping import NDArray, Shape, Float

from ..data.preader import get_yahoofinance_data, get_symbol_closing_price
from .optimize.policy import OptimizedWeightingPolicy
from .helper import align_timestamps_stock_dataframes


class Portfolio:
    def __init__(
            self,
            symbols_nbshares: dict[str, Union[int, float]],    # e.g., symbols_nbshares = {'NVDA': 200, 'AMZN': 101}
            cacheddir: Union[PathLike, str]=None
    ):
        self.symbols_nbshares = symbols_nbshares
        self.cacheddir = cacheddir

    def get_portfolio_value(self, datestr: str) -> float:
        portfolio_value = sum([
            self.symbols_nbshares[symbol] * get_symbol_closing_price(symbol, datestr, cacheddir=self.cacheddir)
            for symbol in self.symbols_nbshares
        ])
        return portfolio_value

    def get_portfolio_values_overtime(
            self,
            startdate: str,
            enddate: str,
            cacheddir: Union[PathLike, str]=None,
            progressbar: bool=False
    ) -> pd.DataFrame:
        logging.debug('Reading financial data...')
        iterator = tqdm(self.symbols_nbshares.keys()) if progressbar else self.symbols_nbshares.keys()
        stocks_data_dfs = [
            get_yahoofinance_data(sym, startdate, enddate, cacheddir=cacheddir)
            for sym in iterator
        ]

        # unify the timestamps columns
        logging.info('Unifying timestamps....')
        stocks_data_dfs = align_timestamps_stock_dataframes(stocks_data_dfs)

        logging.debug('Estimating...')
        max_timearray_ref = 0
        maxlen = max(len(stocks_data_dfs[i]) for i in range(len(stocks_data_dfs)))
        minlen = min(len(stocks_data_dfs[i]) for i in range(len(stocks_data_dfs)))

        if minlen != maxlen:
            logging.warning(f'Not all symbols have data all the way back to {startdate}')
            max_timearray_ref = [i for i in range(len(stocks_data_dfs)) if maxlen == len(stocks_data_dfs[i])][0]
            logging.warning('Symbols not having whole range of data:')
            for i, symbol in enumerate(self.symbols_nbshares):
                thisdflen = len(stocks_data_dfs[i])
                if thisdflen != maxlen:
                    if thisdflen == 0:
                        logging.warning(f'No data for {symbol} for this date range at all.')
                        predf = pd.DataFrame(stocks_data_dfs[max_timearray_ref]['TimeStamp'])
                    else:
                        startdatestr = stocks_data_dfs[i].loc[0, 'TimeStamp'].date().strftime('%Y-%m-%d')
                        logging.warning(f'{symbol}: starting from {startdatestr}')
                        predf = pd.DataFrame(stocks_data_dfs[max_timearray_ref].loc[:(maxlen - thisdflen), 'TimeStamp'])
                    predf[stocks_data_dfs[max_timearray_ref].columns[1:]] = 0
                    predf = predf.append(stocks_data_dfs[i]).reset_index()
                    del predf['index']
                    stocks_data_dfs[i] = predf

        df = pd.DataFrame(stocks_data_dfs[max_timearray_ref]['TimeStamp'].copy())
        df['value'] = sum([
            self.symbols_nbshares[sym] * stocks_data_dfs[i]['Close']
            for i, sym in enumerate(self.symbols_nbshares.keys())
        ])
        return df

    def __add__(self, other: Self) -> Self:
        assert isinstance(other, Portfolio)

        symshares1 = defaultdict(lambda: 0, self.symbols_nbshares)
        symshares2 = defaultdict(lambda: 0, other.symbols_nbshares)
        all_symbols = set(symshares1.keys()).union(symshares2.keys())

        total_symbols_shares = {symbol: symshares2[symbol] + symshares1[symbol] for symbol in all_symbols}
        return Portfolio(total_symbols_shares, cacheddir=self.cacheddir)

    def __sub__(self, other: Self) -> Self:
        assert isinstance(other, Portfolio)

        symshares1 = defaultdict(lambda: 0, self.symbols_nbshares)
        symshares2 = defaultdict(lambda: 0, other.symbols_nbshares)
        all_symbols = set(symshares1.keys()).union(symshares2.keys())

        symbols_diff_shares = {symbol: symshares1[symbol] - symshares2[symbol] for symbol in all_symbols}
        return Portfolio(symbols_diff_shares, cacheddir=self.cacheddir)

    def __eq__(self, other: Self) -> bool:
        assert isinstance(other, Portfolio)

        for symbol, nbshares in self.symbols_nbshares.items():
            if symbol not in other.symbols_nbshares:
                return False
            if other.symbols_nbshares[symbol] != nbshares:
                return False

        return True

    def __ne__(self, other: Self) -> bool:
        return not self.__eq__(other)

    @property
    def portfolio_symbols_nbshares(self) -> dict[str, Union[int, float]]:
        return self.symbols_nbshares

    def roundoff_nbshares(self) -> None:
        for symbol in self.symbols_nbshares:
            nbshares = round(self.symbols_nbshares[symbol])
            self.symbols_nbshares[symbol] = nbshares

    def multiply(self, factor: float) -> None:
        for symbol in self.symbols_nbshares:
            nbshares = self.symbols_nbshares[symbol]
            self.symbols_nbshares[symbol] = nbshares * factor
            
    def __mul__(self, other: dict[str, Union[int, float]]) -> Self:
        assert isinstance(other, int) or isinstance(other, float)
        newshares = {symbol: nbshares*other for symbol, nbshares in self.symbols_nbshares.items()}
        return Portfolio(newshares, cacheddir=self.cacheddir)

    def save_to_json(self, fileobj: TextIOWrapper) -> None:
        json.dump(self.symbols_nbshares, fileobj)

    def dumps_json(self) -> str:
        return json.dumps(self.symbols_nbshares)

    @classmethod
    def load_from_json(
            cls,
            fileobj: TextIOWrapper,
            cacheddir: Union[PathLike, str]=None
    ) -> Self:
        symbols_nbshares = json.load(fileobj)
        return cls(symbols_nbshares, cacheddir=cacheddir)

    @classmethod
    def load_from_dict(
            cls,
            portdict: dict[str, Union[int, float]],
            cacheddir: Union[PathLike, str]=None
    ) -> Self:
        return cls(portdict, cacheddir=cacheddir)


class OptimizedPortfolio(Portfolio):
    def __init__(
            self,
            policy: OptimizedWeightingPolicy,
            totalworth: float,
            presetdate: str,
            cacheddir: Union[PathLike, str]=None
    ):
        super(OptimizedPortfolio, self).__init__({}, cacheddir=cacheddir)
        self.policy = policy
        self.totalworth = totalworth
        self.presetdate = presetdate
        self.compute()

    def compute(self) -> None:
        prices = {
            symbol: get_symbol_closing_price(symbol, self.presetdate, cacheddir=self.cacheddir)
            for symbol in self.policy.symbols
        }
        summary = self.policy.portfolio_summary
        for component in summary['components']:
            symbol = component['symbol']
            component['nbshares'] = component['weight'] * self.totalworth / prices[symbol]
            self.symbols_nbshares[symbol] = component['nbshares']

        self.summary = summary

    @property
    def portfolio_symbols(self) -> list[str]:
        return self.policy.portfolio_symbols

    @property
    def weights(self) -> NDArray[Shape["*"], Float]:
        return self.policy.weights

    @property
    def portfolio_yield(self) -> float:
        return self.policy.portfolio_yield

    @property
    def volatility(self) -> float:
        return self.policy.volatility

    @property
    def correlation_matrix(self) -> NDArray[Shape["*, *"], Float]:
        return self.policy.correlation_matrix

    @property
    def named_correlation_matrix(self) -> pd.DataFrame:
        return self.policy.named_correlation_matrix

    @property
    def portfolio_summary(self) -> dict[str, Any]:
        return self.summary

    def get_portfolio(self) -> Portfolio:
        return Portfolio(self.symbols_nbshares, cacheddir=self.cacheddir)

    def save_to_json(self, fileobj: TextIOWrapper) -> None:
        self.get_portfolio().save_to_json(fileobj)

    def dumps_json(self) -> str:
        return self.get_portfolio().dumps_json()

    @classmethod
    def load_from_json(
            cls,
            fileobj: TextIOWrapper,
            cacheddir: Union[PathLike, str]=None
    ) -> Self:
        raise NotImplementedError('OptimizedPortfolio does not implement loading from json. Use Portfolio to load instead.')

    @classmethod
    def load_from_dict(
            cls,
            portdict: TextIOWrapper,
            cacheddir: Union[PathLike, str]=None
    ) -> Self:
        raise NotImplementedError('OptimizedPortfolio does not implement loading from json. Use Portfolio to load instead.')




import unittest
from math import exp
from datetime import datetime, timedelta

import numpy as np

from finsim.estimate.fit import fit_BlackScholesMerton_model
from finsim.simulation.stock import BlackScholesMertonStockPrices, HestonStockPrices, MertonJumpDiffusionStockPrices


def expected_bsm_stock(S0, r, sigma, t):
    expected_S = S0 * exp(r*t)
    variance_S = S0 * S0 * (exp(sigma*sigma*t)-1) * exp((2*r+sigma*sigma)*t)
    return expected_S, variance_S


class TestStockSimulations(unittest.TestCase):
    def backend_test_BlackScholesMertonStocks(self, S, r, sigma, T, dt, nbsimulations):
        numerical_ts = np.linspace(0, T, int(T // dt) + 1)
        ts = np.array([
            datetime(2020, 1, 1, 0, 0, 0) + timedelta(days=numerical_ts[i])
            for i in range(len(numerical_ts))
        ], dtype='datetime64[s]')

        bsm_simulator = BlackScholesMertonStockPrices(S, r, sigma)
        sarray = bsm_simulator.generate_time_series(T, dt, nbsimulations=nbsimulations)
        rsigma_array = np.array([fit_BlackScholesMerton_model(ts, sarray[i, :], unit='day')
                                 for i in range(nbsimulations)])
        average_r = np.nanmean(rsigma_array[:, 0])
        std_r = np.nanstd(rsigma_array[:, 0])
        average_sigma = np.nanmean(rsigma_array[:, 1])
        std_sigma = np.nanstd(rsigma_array[:, 1])

        self.assertAlmostEqual(float(average_r), r, delta=2.576*std_r)
        self.assertAlmostEqual(float(average_sigma), sigma, delta=2.576*std_sigma)

    def test_BlackScholesMertonStocks(self):
        self.backend_test_BlackScholesMertonStocks(100., 0.04, 0.02, 10, 0.1, 100000)

        for S in np.random.normal(100., 5., size=3):
            for r in np.random.normal(0.1, 0.01, size=3):
                for sigma in np.random.uniform(0.001, 0.05, size=3):
                    self.backend_test_BlackScholesMertonStocks(S, r, sigma, 10, 0.1, 1000)

    def test_HestonStocks(self):
        h_simulator = HestonStockPrices(100, 0.01, 0.01, 0.01, 0.1, 0.0001, 0.2)
        h_stocks = h_simulator.generate_time_series(100, 0.1, nbsimulations=100000)[0]

        self.assertAlmostEqual(
            float(np.nanmean(h_stocks[:, 100])),
            110.52,
            delta=2.576*0.11
        )
        self.assertAlmostEqual(
            float(np.nanmean(h_stocks[:, 250])),
            128.41,
            delta=2.576*0.23
        )
        self.assertAlmostEqual(
            float(np.nanmean(h_stocks[:, 500])),
            164.92,
            delta=2.576*0.38
        )
        self.assertAlmostEqual(
            float(np.nanmean(h_stocks[:, 750])),
            211.80,
            delta=2.576*0.56
        )
        self.assertAlmostEqual(
            float(np.nanmean(h_stocks[:, 999])),
            271.72,
            delta=2.576*0.80
        )

    def test_MertonJumpDiffusionStocks(self):
        mjd_simulator = MertonJumpDiffusionStockPrices(100, 0.01, 0.01, 0.01, 0.01, 0.001)
        mjd_stocks = [
            np.mean(mjd_simulator.generate_time_series(100, 0.1, nbsimulations=1000), axis=0)
            for _ in range(100)
        ]

        self.assertAlmostEqual(
            float(np.mean([stock[250] for stock in mjd_stocks])),
            197.97,
            delta=1.96*6.37
        )
        self.assertAlmostEqual(
            float(np.mean([stock[500] for stock in mjd_stocks])),
            393.35,
            delta=1.96*28.15
        )
        self.assertAlmostEqual(
            float(np.mean([stock[750] for stock in mjd_stocks])),
            773.20,
            delta=1.96*67.73
        )
        self.assertAlmostEqual(
            float(np.mean([stock[999] for stock in mjd_stocks])),
            1523.06,
            delta=1.96*169.68
        )


if __name__ == '__main__':
    unittest.main()




import pandas as pd
import quandl


class QuandlReader:
    def __init__(self, token: str):
        self.token = token

    def get_treasury_real_longterm_rates(self) -> pd.DataFrame:
        return quandl.get("USTREASURY/REALLONGTERM", authtoken=self.token)

    def get_treasury_real_yield_curve_rates(self) -> pd.DataFrame:
        return quandl.get("USTREASURY/REALYIELD", authtoken=self.token)

    def get_treasury_yield_curve_rates(self) -> pd.DataFrame:
        return quandl.get("USTREASURY/YIELD", authtoken=self.token)

    def get_treasury_longterm_rates(self) -> pd.DataFrame:
        return quandl.get("USTREASURY/LONGTERMRATES", authtoken=self.token)




from typing import Literal

import numpy as np
from nptyping import NDArray, Shape, Float

from .native.pymetrics import python_sharpe_ratio, python_mpt_costfunction, python_mpt_entropy_costfunction


def sharpe_ratio(
        weights: NDArray[Shape["*"], Float],
        r: NDArray[Shape["*"], Float],
        cov: NDArray[Shape["*, *"], Float],
        rf: float,
        lowlevellang: Literal['C', 'P']='P'
) -> float:
    if lowlevellang == 'C':
        raise ValueError("Cython fitting is no longer supported!")
    elif lowlevellang == 'P':
        return python_sharpe_ratio(weights, r, cov, rf)
    else:
        raise ValueError('Unknown low-level language: {}. (Should be "P" (Python), or "C" (Cython).)'.format(lowlevellang))


def mpt_costfunction(
        weights: NDArray[Shape["*"], Float],
        r: NDArray[Shape["*"], Float],
        cov: NDArray[Shape["*, *"], Float],
        rf: float,
        lamb: float,
        V0: float=10.,
        lowlevellang: Literal['C', 'P']='P'
) -> float:
    if lowlevellang == 'C':
        raise ValueError("Cython fitting is no longer supported!")
    elif lowlevellang == 'P':
        return python_mpt_costfunction(weights, r, cov, rf, lamb, V0=V0)
    else:
        raise ValueError('Unknown low-level language: {}. (Should be "P" (Python), or "C" (Cython).)'.format(lowlevellang))


def mpt_entropy_costfunction(
        weights: NDArray[Shape["*"], Float],
        r: NDArray[Shape["*"], Float],
        cov: NDArray[Shape["*, *"], Float],
        rf: float,
        lamb0: float,
        lamb1: float,
        V: float=10.,
        lowlevellang: Literal['C', 'P']='C'
) -> float:
    if lowlevellang == 'C':
        raise ValueError("Cython fitting is no longer supported!")
    elif lowlevellang == 'P':
        return python_mpt_entropy_costfunction(weights, r, cov, rf, lamb0, lamb1, V=V)
    else:
        raise ValueError('Unknown low-level language: {}. (Should be "P" (Python), or "C" (Cython).)'.format(lowlevellang))





from datetime import timedelta, datetime
from typing import Union
from os import PathLike

import numpy as np
import pandas as pd

from ..data.preader import get_yahoofinance_data


def get_movingaverage_price_data(
        symbol: str,
        startdate: str,
        enddate: str,
        dayswindow: int,
        cacheddir: Union[PathLike, str]=None
) -> pd.DataFrame:
    # making the days difference calculation
    mastartdate = (datetime.strptime(startdate, '%Y-%m-%d') - timedelta(days=dayswindow)).strftime('%Y-%m-%d')
    df = get_yahoofinance_data(symbol, mastartdate, enddate, cacheddir=cacheddir)
    nbrecords = len(df)
    lastday = df.iloc[nbrecords-1, :]['TimeStamp']
    daysdiff = lastday - df['TimeStamp']
    madf = df.copy()
    madf['DaysDiff'] = daysdiff.apply(lambda dd: dd.days)
    maxdaysdiff = madf.iloc[0]['DaysDiff']
    madf['computema'] = maxdaysdiff - madf['DaysDiff'] > dayswindow

    # calculating moving average
    madf['MA'] = madf.apply(
        lambda row: np.mean(
            madf.loc[(madf['DaysDiff'] >= row['DaysDiff']) & (madf['DaysDiff'] < row['DaysDiff']+dayswindow), 'Close']
        ),
        axis=1
    )

    return madf.loc[madf['computema'], ['TimeStamp', 'MA']]




import numpy as np
import pandas as pd


def align_timestamps_stock_dataframes(
        stocks_data_dfs: list[pd.DataFrame],
        timestamps_as_index: bool=False
) -> pd.DataFrame:
    # unify the timestamps columns
    timestampset = set()
    for stock_df in stocks_data_dfs:
        timestampset = timestampset.union(stock_df['TimeStamp'])
    alltimestamps = sorted(list(timestampset))
    timedf = pd.DataFrame({'AllTime': alltimestamps})

    # wrangle the stock dataframes
    for i, stock_df in enumerate(stocks_data_dfs):
        stock_df = pd.merge(stock_df, timedf, how='right', left_on='TimeStamp', right_on='AllTime').ffill()
        stock_df = stock_df.loc[~np.isnan(stock_df['TimeStamp']), stock_df.columns[:-1]]
        stock_df.loc[:, 'TimeStamp'] = timedf.loc[(len(timedf)-len(stock_df)):, 'AllTime'].to_numpy()
        stock_df = stock_df.reset_index()
        stock_df = stock_df.drop(['index'], axis=1)
        if timestamps_as_index:
            stock_df.index = stock_df['TimeStamp']
        stocks_data_dfs[i] = stock_df

    return stocks_data_dfs


class InsufficientSharesException(Exception):
    pass




import logging
from functools import partial
from itertools import product
from datetime import datetime
from typing import Union, Tuple
from os import PathLike

import numpy as np
import pandas as pd
from scipy.optimize import minimize, OptimizeResult
from tqdm import tqdm
import numba as nb
from nptyping import NDArray, Shape, Float

from .metrics import sharpe_ratio, mpt_costfunction, mpt_entropy_costfunction
from ..helper import align_timestamps_stock_dataframes
from ...data import get_yahoofinance_data
from ...data.preader import get_dividends_df
from ...estimate.fit import fit_multivariate_BlackScholesMerton_model, fit_BlackScholesMerton_model, \
    fit_timeweighted_BlackScholesMerton_model, fit_timeweighted_multivariate_BlackScholesMerton_model


@nb.njit(nb.float64(nb.float64[:], nb.float64, nb.int32))
def getarrayelementminusminvalue(
        array: NDArray[Shape["*"], Float],
        minvalue: float,
        index: int
) -> float:
    return array[index] - minvalue


@nb.njit(nb.float64(nb.float64[:], nb.float64))
def checksumarray(array: NDArray[Shape["*"], Float], total: float) -> float:
    return total - np.sum(array)


def optimized_portfolio_on_sharperatio(
        r: NDArray[Shape["*"], Float],
        cov: NDArray[Shape["*, *"], Float],
        rf: float,
        minweight: float=0.
) -> OptimizeResult:
    func = partial(sharpe_ratio, r=r, cov=cov, rf=rf)
    nbstocks = len(r)
    initialguess = np.repeat(1 / nbstocks, nbstocks)
    constraints = [
        {'type': 'ineq', 'fun': partial(getarrayelementminusminvalue, minvalue=minweight, index=i)}
        for i in range(nbstocks)
    ] + [
        {'type': 'eq', 'fun': partial(checksumarray, total=1.)}
    ] + [
        {'type': 'ineq', 'fun': lambda weights: weights[i]}
        for i in range(len(r))
    ]
    return minimize(
        lambda weights: -func(weights),
        initialguess,
        constraints=constraints
    )


def optimized_portfolio_mpt_costfunction(
        r: NDArray[Shape["*"], Float],
        cov: NDArray[Shape["*, *"], Float],
        rf: float,
        lamb: float,
        V0: float=10.
) -> OptimizeResult:
    func = partial(mpt_costfunction, r=r, cov=cov, rf=rf, lamb=lamb, V0=V0)
    nbstocks = len(r)
    constraints = [
        {'type': 'ineq', 'fun': partial(getarrayelementminusminvalue, minvalue=0., index=i)}
        for i in range(nbstocks+1)
    ] + [
        {'type': 'ineq', 'fun': partial(checksumarray, total=V0)}
    ] + [
        {'type': 'ineq', 'fun': lambda weights: weights[i]}
        for i in range(len(r))
    ]
    initialguess = np.repeat(V0 / (nbstocks+1), nbstocks+1)
    return minimize(
        lambda weights: -func(weights),
        initialguess,
        constraints=constraints
    )


def optimized_portfolio_mpt_entropy_costfunction(
        r: NDArray[Shape["*"], Float],
        cov: NDArray[Shape["*, *"], Float],
        rf: float,
        lamb0: float,
        lamb1: float,
        V: float=10.
) -> OptimizeResult:
    func = partial(mpt_entropy_costfunction, r=r, cov=cov, rf=rf, lamb0=lamb0, lamb1=lamb1, V=V)
    nbstocks = len(r)
    constraints = [
        {'type': 'ineq', 'fun': partial(getarrayelementminusminvalue, minvalue=0., index=i)}
        for i in range(nbstocks+1)
    ] + [
        {'type': 'ineq', 'fun': partial(checksumarray, total=V)}
    ] + [
        {'type': 'ineq', 'fun': lambda weights: weights[i]}
        for i in range(len(r))
    ]
    initialguess = np.repeat(V / (nbstocks + 1), nbstocks + 1)
    return minimize(
        lambda weights: -func(weights),
        initialguess,
        constraints=constraints
    )


def intermediate_wrangle_stock_df_without_dividends(stock_df: pd.DataFrame) -> pd.DataFrame:
    stock_df.loc[:, 'EffVal'] = stock_df['Close'] * 1.
    return stock_df


def intermediate_wrangle_stock_df_with_dividends(stock_df: pd.DataFrame, sym: str) -> pd.DataFrame:
    dividends_df = get_dividends_df(sym)
    dividends_df = dividends_df.rename(columns={'date': 'TimeStamp'})
    dividends_df.loc[:, 'Cash'] = np.cumsum(dividends_df['Dividends'].ravel())
    stock_df.loc[:, 'TimeStamp'] = stock_df['TimeStamp'].map(lambda ts: datetime.strftime(ts, '%Y-%m-%d'))
    stock_df = stock_df.merge(dividends_df, how='left').ffill().fillna(0)
    stock_df.loc[:, 'EffVal'] = stock_df['Close'] + stock_df['Cash']
    stock_df.loc[:, 'TimeStamp'] = stock_df['TimeStamp'].map(lambda ts: datetime.strptime(ts, '%Y-%m-%d'))
    return stock_df


def get_BlackScholesMerton_stocks_estimation(
        symbols: list[str],
        startdate: str,
        enddate: str,
        progressbar: bool=True,
        cacheddir: Union[PathLike, str]=None,
        include_dividends: bool=False
) -> Tuple[NDArray[Shape["*"], Float], NDArray[Shape["*, *"], Float]]:
    logging.info('Reading financial data...')
    symreadingprogress = tqdm(symbols) if progressbar else symbols
    stocks_data_dfs = [
        get_yahoofinance_data(sym, startdate, enddate, cacheddir=cacheddir)
        for sym in symreadingprogress
    ]

    if include_dividends:
        for i, symbol in enumerate(symbols):
            stocks_data_dfs[i] = intermediate_wrangle_stock_df_with_dividends(stocks_data_dfs[i], symbol)
    else:
        for i in range(len(symbols)):
            stocks_data_dfs[i] = intermediate_wrangle_stock_df_without_dividends(stocks_data_dfs[i])

    # unify the timestamps columns
    logging.info('Unifying timestamps....')
    stocks_data_dfs = align_timestamps_stock_dataframes(stocks_data_dfs)

    # calculating length
    logging.info('Estimating...')
    max_timearray_ref = 0
    maxlen = max(len(stocks_data_dfs[i]) for i in range(len(stocks_data_dfs)))
    minlen = min(len(stocks_data_dfs[i]) for i in range(len(stocks_data_dfs)) if len(stocks_data_dfs) > 0)   # exclude those stocks that do not exist
    absent_stocks = {sym for sym, df in zip(symbols, stocks_data_dfs) if len(df) == 0}
    logging.debug(f'{maxlen=}; {minlen=}; absent_stocks: {", ".join(absent_stocks)}')

    # same length, directly compare
    if maxlen == minlen:
        return fit_multivariate_BlackScholesMerton_model(
            stocks_data_dfs[max_timearray_ref]['TimeStamp'].ravel(),
            np.array([
                stocks_data_dfs[i]['EffVal'].ravel()
                for i in range(len(stocks_data_dfs))
            ])
        )
    else:    # maxlen != minlen:
        rarray = np.zeros(len(symbols))
        covmat = np.zeros((len(symbols), len(symbols)))

        for i, stock_df in enumerate(stocks_data_dfs):
            r, sigma = fit_BlackScholesMerton_model(stock_df['TimeStamp'].to_numpy(), stock_df['Close'].to_numpy())
            rarray[i] = r
            covmat[i, i] = sigma*sigma

        for i, j in product(range(len(symbols)), range(len(symbols))):
            stock_df_i = stocks_data_dfs[i]
            stock_df_j = stocks_data_dfs[j]
            smallerlen = min(len(stock_df_i), len(stock_df_j))
            _, cov = fit_multivariate_BlackScholesMerton_model(
                stock_df_i.loc[(len(stock_df_i)-smallerlen):, 'TimeStamp'].to_numpy(),
                np.array([
                    stock_df_i.loc[(len(stock_df_i)-smallerlen):, 'Close'].to_numpy(),
                    stock_df_j.loc[(len(stock_df_j)-smallerlen):, 'Close'].to_numpy()
                ])
            )
            covmat[i, j] = cov[0, 1]
            covmat[j, i] = cov[1, 0]

        return rarray, covmat


def get_stocks_timeweighted_estimation(
        symbols: list[str],
        timeweightdf: pd.DataFrame,
        progressbar: bool=True,
        cacheddir: Union[PathLike, str]=None,
        include_dividends: bool=False
) -> Tuple[NDArray[Shape["*"], Float], NDArray[Shape["*, *"], Float]]:
    logging.info('Parsing weights according to date')
    startdate = timeweightdf['TimeStamp'][0]
    if isinstance(startdate, datetime):
        startdate = startdate.strftime('%Y-%m-%d')
    enddate = timeweightdf['TimeStamp'][len(timeweightdf) - 1]
    if isinstance(enddate, datetime):
        enddate = enddate.strftime('%Y-%m-%d')

    logging.info('Reading financial data...')
    symreadingprogress = tqdm(symbols) if progressbar else symbols
    stocks_data_dfs = [
        get_yahoofinance_data(sym, startdate, enddate, cacheddir=cacheddir)
        for sym in symreadingprogress
    ]

    if include_dividends:
        for i, symbol in enumerate(symbols):
            stocks_data_dfs[i] = intermediate_wrangle_stock_df_with_dividends(stocks_data_dfs[i], symbol)
    else:
        for i in range(len(symbols)):
            stocks_data_dfs[i] = intermediate_wrangle_stock_df_without_dividends(stocks_data_dfs[i])

    # unify the timestamps columns
    logging.info('Unifying timestamps....')
    stocks_data_dfs = align_timestamps_stock_dataframes(stocks_data_dfs)

    # calculating length
    logging.info('Estimating...')
    max_timearray_ref = 0
    maxlen = max(len(stocks_data_dfs[i]) for i in range(len(stocks_data_dfs)))
    minlen = min(len(stocks_data_dfs[i]) for i in range(len(stocks_data_dfs)) if
                 len(stocks_data_dfs) > 0)  # exclude those stocks that do not exist
    absent_stocks = {sym for sym, df in zip(symbols, stocks_data_dfs) if len(df) == 0}
    logging.debug(f'{maxlen=}; {minlen=}; absent_stocks: {", ".join(absent_stocks)}')

    # same length, directly compare
    if maxlen == minlen and maxlen == len(timeweightdf):
        return fit_timeweighted_multivariate_BlackScholesMerton_model(
            np.array(stocks_data_dfs[max_timearray_ref]['TimeStamp']),
            np.array([
                np.array(stocks_data_dfs[i]['EffVal'])
                for i in range(len(stocks_data_dfs))
            ]),
            timeweightdf['weight'].ravel()
        )
    else:
        rarray = np.zeros(len(symbols))
        covmat = np.zeros((len(symbols), len(symbols)))

        for i, stock_df in enumerate(stocks_data_dfs):
            w_stock_df = stock_df.merge(timeweightdf, on='TimeStamp', how='left')
            r, sigma = fit_timeweighted_BlackScholesMerton_model(
                w_stock_df['TimeStamp'].ravel(),
                w_stock_df['Close'].ravel(),
                w_stock_df['weight'].ravel()
            )
            rarray[i] = r
            covmat[i, i] = sigma * sigma

        for i, j in product(range(len(symbols)), range(len(symbols))):
            w_stock_df_i = stocks_data_dfs[i].merge(timeweightdf, on='TimeStamp', how='left')
            w_stock_df_j = stocks_data_dfs[j].merge(timeweightdf, on='TimeStamp', how='left')
            smallerlen = min(len(w_stock_df_i), len(w_stock_df_j))
            _, cov = fit_timeweighted_multivariate_BlackScholesMerton_model(
                w_stock_df_i.loc[(len(w_stock_df_i) - smallerlen):, 'TimeStamp'].ravel(),
                np.array([
                    w_stock_df_i.loc[(len(w_stock_df_i) - smallerlen):, 'Close'].ravel(),
                    w_stock_df_j.loc[(len(w_stock_df_j) - smallerlen):, 'Close'].ravel()
                ]),
                w_stock_df_i.loc[(len(w_stock_df_i) - smallerlen):, 'weight'].ravel()
            )
            covmat[i, j] = cov[0, 1]
            covmat[j, i] = cov[1, 0]

        return rarray, covmat




from datetime import datetime, timedelta
from typing import Union
from pathlib import Path
from os import PathLike
import logging
from time import sleep
from functools import lru_cache
import threading
import traceback

import pandas as pd
import tables
import yfinance as yf
import tables as tb
from tqdm import tqdm


def extract_online_yahoofinance_data(symbol: str, startdate: str, enddate: str) -> pd.DataFrame:
    try:
        df = yf.download(
            symbol,
            start=datetime.strptime(startdate, '%Y-%m-%d'),
            end=datetime.strptime(enddate, '%Y-%m-%d') + timedelta(days=1),
            auto_adjust=False
        )
    except:
        logging.warning(f'Symbol {symbol} does not exist between {startdate} and {enddate}.')
        return pd.DataFrame({
            'TimeStamp': [],
            'High': [],
            'Low': [],
            'Open': [],
            'Close': [],
            'Adj Close': [],
            'Volume': [],
        })

    oricols = df.columns
    if isinstance(oricols, pd.core.indexes.multi.MultiIndex):
        df.columns = [col[0] for col in oricols]
        oricols = df.columns
    df['TimeStamp'] = df.index
    df = df.loc[:, ['TimeStamp'] + list(oricols)].reset_index()

    return df


def extract_batch_online_yahoofinance_data(
        symbols: list[str],
        startdate: str,
        enddate: str,
        threads: bool=True
) -> pd.DataFrame:
    combined_df = yf.download(
        ' '.join(symbols),
        start=datetime.strptime(startdate, '%Y-%m-%d'),
        end=datetime.strptime(enddate, '%Y-%m-%d') + timedelta(days=1),
        group_by='ticker',
        threads=threads
    )

    dataframes = {}
    for symbol in symbols:
        try:
            df = combined_df[symbol].copy()
            oricols = df.columns
            df['TimeStamp'] = df.index
            df = df.loc[:, ['TimeStamp'] + list(oricols)]
            dataframes[symbol] = df
        except:
            dataframes[symbol] = pd.DataFrame({
                'TimeStamp': [],
                'High': [],
                'Low': [],
                'Open': [],
                'Close': [],
                'Adj Close': [],
                'Volume': [],
            })

    return dataframes


# yahoo reader with local cache
METATABLE_FILENAME = 'SYMBOL_CACHE.h5'
METATABLE_ROWDES = {
    'symbol': tb.StringCol(shape=(), itemsize=10, dflt='', pos=0),
    'query_startdate': tb.StringCol(shape=(), itemsize=10, dflt='0000-00-00', pos=1),
    'query_enddate': tb.StringCol(shape=(), itemsize=10, dflt='0000-00-00', pos=2),
    'data_startdate': tb.StringCol(shape=(), itemsize=10, dflt='0000-00-00', pos=3),
    'data_enddate': tb.StringCol(shape=(), itemsize=10, dflt='0000-00-00', pos=4)
}


def get_yahoofinance_data(
        symbol: str,
        startdate: str,
        enddate: str,
        cacheddir: Union[str, PathLike]=None
) -> pd.DataFrame:
    if cacheddir is None:
        return extract_online_yahoofinance_data(symbol, startdate, enddate)

    if isinstance(cacheddir, str):
        cacheddir = Path(cacheddir)

    if not cacheddir.is_dir():
        logging.info(f'Creating directory: {cacheddir.as_posix()}')
        cacheddir.mkdir()

    cached_metafile_path = cacheddir / METATABLE_FILENAME
    if not cached_metafile_path.exists():
        logging.info(f'Creating file: {cached_metafile_path.as_posix()}')
        metatable_h5file = tb.open_file(cached_metafile_path.as_posix(), 'w')
        table = metatable_h5file.create_table('/', 'metatable', METATABLE_ROWDES, title='metatable')
    else:
        metatable_h5file = tb.open_file(cached_metafile_path.as_posix(), 'r+')
        table = metatable_h5file.root.metatable

    preexist = False
    for row in table.where('symbol=="{}"'.format(symbol)):
        preexist = True
        if row['query_startdate'].decode('utf-8') <= startdate and row['query_enddate'].decode('utf-8') >= enddate:
            df = pd.read_hdf((cacheddir / f"{symbol}.h5").as_posix(), key='yahoodata')
            if len(df) > 0:
                df = df[(df['TimeStamp'] >= startdate) & (df['TimeStamp'] <= enddate)]
            metatable_h5file.close()
            df = df[~df['Close'].isna()]
            return df

    df = extract_online_yahoofinance_data(symbol, startdate, enddate)
    logging.debug(f'Caching data for {symbol} from {startdate} to {enddate}')
    df.to_hdf(cacheddir / f"{symbol}.h5", key='yahoodata')

    if preexist:
        logging.debug(f'Updating symbol {symbol} in metatable.')
        for row in table.where('symbol=="{}"'.format(symbol)):
            row['query_startdate'] = startdate
            row['query_enddate'] = enddate
            if len(df) > 0:
                row['data_startdate'] = datetime.strftime(df['TimeStamp'].to_list()[0].date(), '%Y-%m-%d')
                row['data_enddate'] = datetime.strftime(df['TimeStamp'].to_list()[-1].date(), '%Y-%m-%d')
            else:
                row['data_startdate'] = '0000-00-00'
                row['data_enddate'] = ' 0000-00-00'
            row.update()
    else:
        logging.debug(f'Creating symbol {symbol} in metatable')
        newrow = table.row
        newrow['symbol'] = symbol
        newrow['query_startdate'] = startdate
        newrow['query_enddate'] = enddate
        if len(df) > 0:
            newrow['data_startdate'] = datetime.strftime(df['TimeStamp'].to_list()[0].date(), '%Y-%m-%d')
            newrow['data_enddate'] = datetime.strftime(df['TimeStamp'].to_list()[-1].date(), '%Y-%m-%d')
        else:
            newrow['data_startdate'] = '0000-00-00'
            newrow['data_enddate'] = '0000-00-00'
        newrow.append()

    table.flush()
    metatable_h5file.close()

    df = df[~df['Close'].isna()]

    return df


@lru_cache(maxsize=256)
def get_symbol_closing_price(
        symbol: str,
        datestr: str,
        epsilon: float=1e-10,
        cacheddir: Union[PathLike, str]=None,
        backtrack: bool=False
) -> float:
    df = get_yahoofinance_data(symbol, datestr, datestr, cacheddir=cacheddir)
    if len(df) == 0:
        if backtrack:
            prevdatestr = datetime.strftime(datetime.strptime(datestr, '%Y-%m-%d') - timedelta(days=1), '%Y-%m-%d')
            return get_symbol_closing_price(symbol, prevdatestr, epsilon=epsilon, cacheddir=cacheddir, backtrack=True)
        else:
            raise IndexError('Price not found!')
    else:
        return df['Close'][0]


def finding_missing_symbols_in_cache(
        symbols: list[str],
        startdate: str,
        enddate: str,
        cacheddir: Union[str, Path]
) -> list[str]:
    if isinstance(cacheddir, str):
        cacheddir = Path(cacheddir)

    cached_metafile_path = cacheddir / METATABLE_FILENAME
    if not cached_metafile_path.exists():
        return symbols

    # in table
    metatable = pd.read_hdf(cached_metafile_path, 'metatable')
    existing_within_range_symbols = list(
        metatable['symbol'][
            (metatable['query_startdate'] <= startdate) & (metatable['query_enddate'] >= enddate)
        ]
    )
    if logging.root.level >= logging.DEBUG:
        logging.debug('exisiting within range symbols')
        for symbol in existing_within_range_symbols:
            logging.debug('\t{}'.format(symbol))

    # check what are in the cached directories
    existing_symbols = [
        symbolfilepath.stem
        for symbolfilepath in cacheddir.glob("*.h5")
        if symbolfilepath is not cached_metafile_path
    ]
    if logging.root.level >= logging.DEBUG:
        logging.debug('exisiting symbols')
        for symbol in existing_symbols:
            logging.debug('\t{}'.format(symbol))

    existing_valid_symbols = set(existing_within_range_symbols) & set(existing_symbols)

    return sorted(list(set(symbols) - set(existing_valid_symbols)))


def dataframe_to_hdf(df: pd.DataFrame, filepath: Union[PathLike, str], key: str) -> None:
    df.to_hdf(filepath, key=key)


def generating_cached_yahoofinance_data(
        symbols: list[str],
        startdate: str,
        enddate: str,
        cacheddir: Union[PathLike, str],
        slicebatch: int=50,
        waittime: int=1,
        yfinance_multithreads: bool=False,
        io_multithreads: bool=False
) -> None:
    if isinstance(cacheddir, str):
        cacheddir = Path(cacheddir)
    cached_metafile_path = cacheddir / METATABLE_FILENAME
    tocache_symbols = finding_missing_symbols_in_cache(symbols, startdate, enddate, cacheddir)

    logging.info(f'Total number of symbols: {len(symbols)}')
    logging.info(f'Total number of symbols needed to cache: {len(tocache_symbols)}')
    if not cacheddir.is_dir():
        logging.info(f'Creating directory: {cacheddir.as_posix()}')
        cacheddir.mkdir()
    if not cached_metafile_path.exists():
        logging.info(f'Creating file: {cached_metafile_path.as_posix()}')
        metatable_h5file = tb.open_file(cached_metafile_path.as_posix(), 'w')
        table = metatable_h5file.create_table('/', 'metatable', METATABLE_ROWDES, title='metatable')
    else:
        metatable_h5file = tb.open_file(cached_metafile_path.as_posix(), 'r+')
        table = metatable_h5file.root.metatable

    nbsymbols = len(tocache_symbols)
    for startidx in tqdm(range(0, nbsymbols, slicebatch)):
        success = False
        while not success:
            try:
                dataframes = extract_batch_online_yahoofinance_data(
                    tocache_symbols[startidx:min(startidx + slicebatch, nbsymbols)],
                    startdate,
                    enddate,
                    threads=yfinance_multithreads
                )
                success = True
            except:
                sleep(waittime)

        writing_threads = []
        for symbol in dataframes:
            df = dataframes[symbol]
            df = df[~df['Close'].isna()]
            if len(df) > 0:
                thissymbol_startdate = datetime.strftime(df['TimeStamp'].to_list()[0].date(), '%Y-%m-%d')
                thissymbol_enddate = datetime.strftime(df['TimeStamp'].to_list()[-1].date(), '%Y-%m-%d')
            else:
                thissymbol_startdate = '0000-00-00'
                thissymbol_enddate = '0000-00-00'

            logging.debug(f'Caching data for {symbol} from {startdate} to {enddate}')
            if not io_multithreads:
                dataframe_to_hdf(df, cacheddir / f'{symbol}.h5', key='yahoodata')
            else:
                thread = threading.Thread(
                    target=dataframe_to_hdf,
                    args=(df, cacheddir / f'{symbol}.h5', 'yahoodata')
                )
                thread.start()
                writing_threads.append(thread)

                try:
                    logging.debug(f'Creating symbol {symbol} in metatable')
                    newrow = table.row
                    newrow['symbol'] = symbol
                    newrow['query_startdate'] = startdate
                    newrow['query_enddate'] = enddate
                    newrow['data_startdate'] = thissymbol_startdate
                    newrow['data_enddate'] = thissymbol_enddate
                    newrow.append()

                except tables.HDF5ExtError as e:
                    logging.error(f'Cannot append record for symbol {symbol}')
                    traceback.print_exc()
                    continue

            table.flush()

        if io_multithreads:
            for thread in writing_threads:
                thread.join()

    metatable_h5file.close()


@lru_cache(maxsize=30)
def get_dividends_df(symbol: str) -> pd.DataFrame:
    ticker = yf.Ticker(symbol)
    df = pd.DataFrame(ticker.dividends)
    df['TimeStamp'] = df.index.map(lambda item: datetime.strftime(item, '%Y-%m-%d'))
    df = df.loc[:, ['TimeStamp', 'Dividends']]
    return df




from itertools import product
from abc import ABC, abstractmethod
from typing import Any

import numpy as np
import pandas as pd
from nptyping import NDArray, Shape, Float

from .numerics import optimized_portfolio_on_sharperatio, optimized_portfolio_mpt_costfunction, optimized_portfolio_mpt_entropy_costfunction


def mat_to_list(mat: NDArray[Shape["*, *"], Float]) -> list[list[float]]:
    return [
        [
            float(mat[i, j]) for j in range(mat.shape[1])
        ]
        for i in range(mat.shape[0])
    ]


class OptimizedWeightingPolicy(ABC):
    def __init__(
            self,
            rf: float,
            r: NDArray[Shape["*"], Float]=None,
            cov: NDArray[Shape["*, *"], Float]=None,
            symbols: list[str]=None
    ):
        self.rf = rf

        assert len(r) == cov.shape[0]
        assert cov.shape[0] == cov.shape[1]
        if symbols is not None:
            assert len(r) == len(symbols)

        self.r = r
        self.cov = cov
        self.symbols = symbols if symbols is not None else list(range(len(r)))

        self.optimized = False

    @abstractmethod
    def optimize(
            self,
            r: NDArray[Shape["*"], Float],
            cov: NDArray[Shape["*, *"], Float],
            symbols: list[str]=None
    ) -> None:
        pass

    @property
    def portfolio_symbols(self) -> list[str]:
        return self.symbols

    @property
    @abstractmethod
    def weights(self) -> NDArray[Shape["*"], Float]:
        pass

    @property
    @abstractmethod
    def portfolio_yield(self) -> float:
        pass

    @property
    @abstractmethod
    def volatility(self) -> float:
        pass

    @property
    def correlation_matrix(self) -> NDArray[Shape["*, *"], Float]:
        corr = np.zeros(self.cov.shape)
        for i, j in product(range(self.cov.shape[0]), range(self.cov.shape[1])):
            corr[i, j] = self.cov[i, j] / np.sqrt(self.cov[i, i] * self.cov[j, j])
        return corr

    @property
    def named_correlation_matrix(self) -> pd.DataFrame:
        corrdf = pd.DataFrame(self.correlation_matrix,
                              columns=self.symbols,
                              index=self.symbols)
        return corrdf

    @property
    def portfolio_summary(self) -> dict[str, Any]:
        summary = {
            'components': [
                {
                    'symbol': sym,
                    'yield': self.r[i],
                    'weight': self.weights[i],
                    'volatility': np.sqrt(self.cov[i, i])
                }
                for i, sym in enumerate(self.symbols)
            ],
            'yield': self.optimized_portfolio_yield,
            'volatility': self.optimized_volatility,
            'correlation': self.correlation_matrix
        }
        return summary

    @property
    def policytype(self) -> str:
        return 'AbstractOptimizedWeightingPolicy'


class OptimizedWeightingPolicyUsingMPTSharpeRatio(OptimizedWeightingPolicy):
    def __init__(
            self,
            rf: float,
            r: NDArray[Shape["*"], Float]=None,
            cov: NDArray[Shape["*, *"], Float]=None,
            symbols: list[str]=None,
            minweight: float=0.
    ):
        super(OptimizedWeightingPolicyUsingMPTSharpeRatio, self).__init__(rf, r=r, cov=cov, symbols=symbols)
        self.minweight = minweight
        if r is not None and cov is not None:
            self.optimize(r, cov, symbols=symbols)

    def optimize(
            self,
            r: NDArray[Shape["*"], Float],
            cov: NDArray[Shape["*, *"], Float],
            symbols: list[str]=None
    ) -> None:
        super(OptimizedWeightingPolicyUsingMPTSharpeRatio, self).optimize(r, cov, symbols=symbols)
        self.optimized_sol = optimized_portfolio_on_sharperatio(r, cov, self.rf, minweight=self.minweight)
        self.optimized = True

        self.optimized_weights = self.optimized_sol.x
        self.optimized_sharpe_ratio = -self.optimized_sol.fun
        self.optimized_portfolio_yield = np.sum(self.optimized_weights * self.r)
        sqweights = np.dot(
            np.expand_dims(self.optimized_weights, axis=1),
            np.expand_dims(self.optimized_weights, axis=0)
        )
        self.optimized_volatility = np.sqrt(np.sum(sqweights * self.cov))

    @property
    def weights(self) -> NDArray[Shape["*"], Float]:
        return self.optimized_weights

    @property
    def portfolio_yield(self) -> float:
        return self.optimized_portfolio_yield

    @property
    def volatility(self) -> float:
        return self.optimized_volatility

    @property
    def sharpe_ratio(self) -> float:
        return self.optimized_sharpe_ratio

    @property
    def portfolio_summary(self) -> dict[str, Any]:
        summary = super(OptimizedWeightingPolicyUsingMPTSharpeRatio, self).portfolio_summary
        summary['sharpe_ratio'] = self.optimized_sharpe_ratio
        return summary

    @property
    def policytype(self) -> str:
        return 'OptimizedWeightingPolicyUsingMPTSharpeRatio'


class OptimizedWeightingPolicyUsingMPTCostFunction(OptimizedWeightingPolicy):
    def __init__(
            self,
            rf: float,
            r: NDArray[Shape["*"], Float],
            cov: NDArray[Shape["*, *"], Float],
            symbols: list[str]=None,
            lamb: float=None,
            V0: float=10.0
    ):
        super(OptimizedWeightingPolicyUsingMPTCostFunction, self).__init__(rf, r=r, cov=cov, symbols=symbols)
        self.lamb = lamb
        self.V0 = V0

        if r is not None and cov is not None:
            self.optimize(r, cov, symbols=symbols)

    def optimize(
            self,
            r: NDArray[Shape["*"], Float],
            cov: NDArray[Shape["*, *"], Float],
            symbols: list[str]=None
    ) -> None:
        super(OptimizedWeightingPolicyUsingMPTCostFunction, self).optimize(r, cov, symbols=symbols)
        self.optimized_sol = optimized_portfolio_mpt_costfunction(r, cov, self.rf, self.lamb, V0=self.V0)
        self.optimized = True

        self.optimized_unnormalized_weights = self.optimized_sol.x
        self.optimized_weights = self.optimized_unnormalized_weights[:-1] / np.sum(self.optimized_unnormalized_weights[:-1])
        self.optimized_costfunction = -self.optimized_sol.fun
        self.optimized_portfolio_yield = np.sum(self.optimized_weights * self.r)
        sqweights = np.dot(
            np.expand_dims(self.optimized_weights, axis=1),
            np.expand_dims(self.optimized_weights, axis=0)
        )
        self.optimized_volatility = np.sqrt(np.sum(sqweights * self.cov))

    @property
    def weights(self) -> NDArray[Shape["*"], Float]:
        return self.optimized_weights

    @property
    def portfolio_yield(self) -> float:
        return self.optimized_portfolio_yield

    @property
    def volatility(self) -> float:
        return self.optimized_volatility

    @property
    def mpt_costfunction(self) -> float:
        return self.optimized_costfunction

    @property
    def portfolio_summary(self) -> dict[str, Any]:
        summary = super(OptimizedWeightingPolicyUsingMPTCostFunction, self).portfolio_summary
        summary['mpt_costfunction'] = self.mpt_costfunction
        summary['V0'] = self.V0
        summary['lamb'] = self.lamb
        return summary

    @property
    def policytype(self) -> str:
        return 'OptimizedWeightingPolicyUsingMPTCostFunction'


class OptimizedWeightingPolicyUsingMPTEntropyCostFunction(OptimizedWeightingPolicy):
    def __init__(
            self,
            rf: float,
            r: NDArray[Shape["*"], Float]=None,
            cov: NDArray[Shape["*, *"], Float]=None,
            symbols: list[str]=None,
            lamb0: float=None,
            lamb1: float=None,
            V: float=10.0
    ):
        super(OptimizedWeightingPolicyUsingMPTEntropyCostFunction, self).__init__(rf, r=r, cov=cov, symbols=symbols)
        self.lamb0 = lamb0
        self.lamb1 = lamb1
        self.V = V

        if r is not None and cov is not None:
            self.optimize(r, cov, symbols=symbols)

    def optimize(
            self,
            r: NDArray[Shape["*"], Float],
            cov: NDArray[Shape["*, *"], Float],
            symbols: list[str]=None
    ) -> None:
        super(OptimizedWeightingPolicyUsingMPTEntropyCostFunction, self).optimize(r, cov, symbols=symbols)
        self.optimized_sol = optimized_portfolio_mpt_entropy_costfunction(r, cov, self.rf, self.lamb0, self.lamb1, V=self.V)
        self.optimized = True

        self.optimized_unnormalized_weights = self.optimized_sol.x
        self.optimized_weights = self.optimized_unnormalized_weights[:-1] / np.sum(self.optimized_unnormalized_weights[:-1])
        self.optimized_costfunction = -self.optimized_sol.fun
        self.optimized_portfolio_yield = np.sum(self.optimized_weights * self.r)
        sqweights = np.dot(
            np.expand_dims(self.optimized_weights, axis=1),
            np.expand_dims(self.optimized_weights, axis=0)
        )
        self.optimized_volatility = np.sqrt(np.sum(sqweights * self.cov))

    @property
    def weights(self) -> NDArray[Shape["*"], Float]:
        return self.optimized_weights

    @property
    def portfolio_yield(self) -> float:
        return self.optimized_portfolio_yield

    @property
    def volatility(self) -> float:
        return self.optimized_volatility

    @property
    def mpt_entropy_costfunction(self) -> float:
        return self.optimized_costfunction

    @property
    def portfolio_summary(self) -> dict[str, Any]:
        summary = super(OptimizedWeightingPolicyUsingMPTEntropyCostFunction, self).portfolio_summary
        summary['mpt_entropy_costfunction'] = self.mpt_entropy_costfunction
        summary['V'] = self.V
        summary['lamb0'] = self.lamb0
        summary['lamb1'] = self.lamb1
        return summary

    @property
    def policytype(self) -> str:
        return 'OptimizedWeightingPolicyUsingMPTEntropyCostFunction'




from typing import Tuple

import numpy as np
import numba as nb
from nptyping import NDArray, Shape, Float


@nb.njit
def python_fit_BlackScholesMerton_model(
        ts: NDArray[Shape["*"], Float],
        prices: NDArray[Shape["*"], Float]
) -> Tuple[float, float]:
    dlogS = np.log(prices[1:] / prices[:-1])
    dt = ts[1:] - ts[:-1]

    r = np.mean(dlogS / dt)
    sigma = np.std(dlogS / np.sqrt(dt))

    return r, sigma


@nb.njit
def python_fit_multivariate_BlackScholesMerton_model(
        ts: NDArray[Shape["*"], Float],
        multiprices: NDArray[Shape["*, *"], Float]
) -> Tuple[NDArray[Shape["*"], Float], NDArray[Shape["*, *"], Float]]:
    dlogS = np.log(multiprices[:, 1:] / multiprices[:, :-1])
    dt = ts[1:] - ts[:-1]

    r = np.zeros(multiprices.shape[0])
    for i in range(multiprices.shape[0]):
        r[i] = np.mean((dlogS[i, :] / dt))
    cov = np.cov(dlogS / np.sqrt(dt), bias=True)

    return r, cov




from abc import ABC, abstractmethod
from math import log, exp

import numpy as np
from nptyping import NDArray, Shape, Float


class AbstractStochasticValue(ABC):
    @abstractmethod
    def generate_time_series(self, T: float, dt: float, nbsimulations: int=1):
        pass


class BlackScholesMertonStockPrices(AbstractStochasticValue):
    def __init__(self, S0: float, r: float, sigma: float):
        self.S0 = S0
        self.r = r
        self.sigma = sigma

        self.logS0 = log(S0)

    def generate_time_series(self, T, dt, nbsimulations=1) -> NDArray[Shape["*"], Float]:
        nbtimesteps = int(T // dt) + 1
        z = np.random.normal(size=(nbsimulations, nbtimesteps))
        logS = np.zeros((nbsimulations, nbtimesteps))
        logS[:, 0] = self.logS0
        for i in range(1, nbtimesteps):
            logS[:, i] = logS[:, i-1] + \
                         (self.r - 0.5 * self.sigma * self.sigma) * dt + \
                         self.sigma * z[:, i] * np.sqrt(dt)
        return np.exp(logS)


class SquareRootDiffusionProcesses(AbstractStochasticValue):
    def __init__(self, x0: float, theta: float, kappa: float, sigma: float):
        self.x0 = x0
        self.theta = theta
        self.kappa = kappa
        self.sigma = sigma

    def generate_time_series(self, T: float, dt: float, nbsimulations: int=1) -> NDArray[Shape["*"], Float]:
        nbtimesteps = int(T // dt) + 1
        z = np.random.normal(size=(nbsimulations, nbtimesteps))
        xarray = np.zeros((nbsimulations, nbtimesteps))
        xarray[:, 0] = self.x0
        for i in range(1, nbtimesteps):
            xarray[:, i] = xarray[:, i - 1] + \
                           self.kappa * (self.theta - xarray[:, i - 1]) * dt + \
                           self.sigma * np.sqrt(xarray[:, i - 1]) * z[:, i] * np.sqrt(dt)
        return xarray


class HestonStockPrices(AbstractStochasticValue):
    def __init__(self, S0: float, r: float, v0: float, theta: float, kappa: float, sigma_v: float, rho: float):
        self.S0 = S0
        self.r = r
        self.v0 = v0
        self.theta = theta
        self.kappa = kappa
        self.sigma_v = sigma_v
        self.rho = rho

        self.logS0 = log(self.S0)
        self.rho = np.array([[1., self.rho], [self.rho, 1.]])

    def generate_time_series(self, T: float, dt: float, nbsimulations: int=1) -> NDArray[Shape["*"], Float]:
        nbtimesteps = int(T // dt) + 1

        # generate correlated random numbers
        Z = np.random.multivariate_normal((0., 0.), self.rho, size=(nbsimulations, nbtimesteps))
        z1 = Z[:, :, 0]
        z2 = Z[:, :, 1]

        # stochastic volatility
        v = np.zeros((nbsimulations, nbtimesteps))
        v[:, 0] = self.v0
        for i in range(1, nbtimesteps):
            v[:, i] = v[:, i - 1] + \
                      self.kappa * (self.theta - v[:, i - 1]) * dt + \
                      self.sigma_v * np.sqrt(v[:, i - 1]) * z2[:, i] * np.sqrt(dt)

        # stock price
        logS = np.zeros((nbsimulations, nbtimesteps))
        logS[:, 0] = self.logS0
        for i in range(1, nbtimesteps):
            logS[:, i] = logS[:, i-1] + \
                         (self.r - 0.5 * v[:, i]) * dt + \
                         np.sqrt(v[:, i]) * z1[:, i] * np.sqrt(dt)

        return np.exp(logS), v


class MertonJumpDiffusionStockPrices(AbstractStochasticValue):
    def __init__(self, S0: float, r: float, sigma: float, mu: float, lamb: float, delta: float):
        self.S0 = S0
        self.r = r
        self.sigma = sigma
        self.mu = mu
        self.lamb = lamb
        self.delta = delta

        self.logS0 = log(self.S0)

    def generate_time_series(self, T: float, dt: float, nbsimulations: int=1) -> NDArray[Shape["*"], Float]:
        nbtimesteps = int(T // dt) + 1
        z1 = np.random.normal(size=(nbsimulations, nbtimesteps))
        logS = np.zeros((nbsimulations, nbtimesteps))
        logS[:, 0] = self.logS0
        rJ = self.lamb * (exp(self.mu+0.5*self.delta*self.delta) - 1)
        nbjumps = np.random.poisson(self.lamb*dt, (nbsimulations, nbtimesteps))
        jumpmagnitudes = np.random.lognormal(
            log(1+self.mu)-0.5*self.delta*self.delta,
            self.delta,
            size=(nbsimulations, nbtimesteps)
        )
        for i in range(1, nbtimesteps):
            logS[:, i] = logS[:, i-1] + \
                         (self.r - rJ - 0.5 * self.sigma * self.sigma) * dt + \
                         self.sigma * z1[:, i] * np.sqrt(dt) + \
                         jumpmagnitudes[:, i] * nbjumps[:, i]
        return np.exp(logS)
